# Hash comparison

This repository provides additional benchmarks for different hashing algorithms and hash table implementations for
the [Go VSL compiler](https://github.com/hhramberg/go-vslc). The purpose of this repository is to separate compiler
source code from hash table benchmarking source code. The results of these benchmarks and review will influence the
choice of hashing table that is the foundation of the VSL compiler's symbol table.

## Preface

The [2021 edition](https://folk.ntnu.no/michaeng/tdt4205_21/)
of the [TDT4205 Compiler Construction course](https://www.ntnu.edu/studies/courses/TDT4205)
uses the [CRC32](https://en.wikipedia.org/wiki/Cyclic_redundancy_check#CRC-32_algorithm)
hashing algorithm for generating an index for hash table lookups for the compiler symbol table. It is my intention to
benchmark the CRC32 hash table implementation and see how it compares to the built-in Go hash map. If the tests suggests
that the Go built-in hash map is nearly as good, or better, it would be prudent to tell students to utilise the built-in
hash map for any future editions of the course. This would save students time from understanding the exact
implementation of the hash table and its
`insert` and `lookup` functions. This will give them more time to focus on the core principles of compiler construction.

## Candidates

- **CRC32 port**: a direct port from C source code to Go located in `src/crc32.go`. Referred to as `crc32p`.
- **CRC32 re-implementation**: a Go hash table implementation using the CRC32 hashing algorithm located
  in `src/crc32_new.go`. Referred to as `crc32re`.
- **CRC32 optimised**: improved version of `crc32re` which only accepts string keys and `Symbol` data type. Uses bucket
  tables, like the Go hashmap, and not linked lists, like the previous two. Referred to as `crc32opt`.
- **DJB2 optimised*: same as CRC32 optimised, but with DJB2 hashing algorithm.
- **Go hash map**: The default built-in hash map of the Go language. Referred to as `goh`.

The `crc32p` was ported form the original C source code, as provided by the professor in the 2021 edition of the TDT4205
course. No optimisations were made. This means that functions still requires `key length` even though Go has string
lengths by default. The `void` data type was replaced by `interface`. The `free` function sets an `interface`
pointer to `nil`. The `memcmp` function was implemented to comply with the definition of the C function. The `copy`
function is built-in in Go, but serves the same purpose, except that it accepts slices and not arbitrary memory
pointers.

The `crc32re` was built from the ground up to be convivial in the Go language environment. The CRC32 hash function is
the same as in `crc32p`. The hash table consists of a slice of `hElements`, of a given size. `hElement` has the
following fields:

```go
type hElement struct {
    key  string
    val  interface{}
    next *hElement
}
```

where `key` is the distinct string key of the element, `val` is the arbitrary data held by the element and `next` is a
pointer to another `hElement` such that the table can be extended to multiple single linked lists in case of collisions.

Where `crc32p` implements a *queue* for the linked list, the `crc32re` implements a *stack*, where elements that are
assigned the same index in the underlying slice are appended to the linked list, not pushing existing elements further
down the list.

The `crc32opt` is very much like `crc32re`, but differs in that it uses a linked list of bucket tables of size 8, rather
than a linked list of elements. Additionally, it accepts only `Symbol` data type as data values. It is expected that
this candidate's *remove* operation is slow, because it has to pop the last element in the bucket stack to insert at the
position of the removed element. This is not crucial for the performance of the hash table as a symbol table, because
removal of symbols rarely happens, if it happens at all.

Finally, there is the `djb2opt` candidate, which is exactly the same as `crc32opt` but it uses the *DJB2* hashing
algorithm instead of CRC32.

## Methodology

Go has built-in support for benchmarking using the `go test` application. Benchmarking dynamically adjusts run periods,
but a run period is
[minimum 1 second](https://dave.cheney.net/2013/06/30/how-to-write-benchmarks-in-go). For operations that mutates
(*insert* and *remove*) the hash table, a
[helper function](https://pkg.go.dev/testing#pkg-overview) creates a new hash table for every iteration of the
benchmark.

There are three benchmarks that tests the performance of the three key functions:

- Insert
- Lookup
- Remove

All tests and benchmarks are located in the `src/crc32_test.go` file. The global variable array `symbols` holds test
data to be inserted into the hash tables.

## Experimental setup

The benchmarks were executed on a Lenovo T430 laptop. The specific model number is Lenovo ThinkPad 2347GU8.

|Component|Description|
|---|---|
|CPU|Intel i5-3320M|
|Memory|8 GB|
|Storage|128 GB SSD|
|Operating system|Manjaro Linux 21.2.1 (Qonos)|

*The laptop was connected to mains power to allow it to draw as much power as needed during the benchmarks*.

## Results

As a reminder the CRC32 port is named `crc32p`, the re-implementation of the former is named `crc32re`, `crc32opt` is
the optimised re-implementation and the built-in Go hash map is named `goh`. **b.N** is the built-in variable that
controls the number of iterations a benchmark function is executed. More on Go benchmarking
[here](https://pkg.go.dev/testing#pkg-overview). **n** is the size of underlying slice supporting the `crc32re`
and `crc32p` hash tables and the initialised size of the `goh` hash map. There were 10 test entries that were inserted
into the hash table.

|Candidate|Operation|n|b.N|Result (ns/op)|
|---|---|---|---|---|
|crc32p|Insert|10|339699|3755|
|crc32p|Lookup|10|2372703|475|
|crc32p|Remove|10|252591|5153|
|crc32re|Insert|10|784000|1605|
|crc32re|Lookup|10|2409684|472|
|crc32re|Remove|10|405198|3532|
|crc32opt|Insert|10|1177317|976|
|crc32opt|Lookup|10|2560135|463|
|crc32opt|Remove|10|312448|4679|
|djb2opt|Insert|10|1529737|790|
|djb2opt|Lookup|10|2563114|469|
|djb2opt|Remove|10|290876|4262|
|goh|Insert|10|1346206|872|
|goh|Lookup|10|5438748|204|
|goh|Remove|10|462085|2431|

*Results: ns/op means [nanoseconds per loop](https://pkg.go.dev/testing#pkg-overview). Lower is better.*

### Insert

|Candidate|crc32p|crc32re|crc32opt|djb2opt|goh|
|---|---|---|---|---|---|
|ns/op|3755|1605|976|790|872|

### Lookup

|Candidate|crc32p|crc32re|crc32opt|djb2opt|goh|
|---|---|---|---|---|---|
|ns/op|475|472|463|469|204|

### Remove

|Candidate|crc32p|crc32re|crc32opt|djb2opt|goh|
|---|---|---|---|---|---|
|ns/op|5153|3532|4679|4262|2431|

### Distribution factor

Distribution factor measures a fraction of how many buckets of the underlying slice have buckets with data in them. The
idea is to measure how well the hashing algorithm distributes elements. An ideal distribution would be 1.0. An
unfavourable distribution will approximate to zero, as multiple elements congest on a single or very few buckets.
**n** is the number of buckets. **n** was set to 10. Since all three CRC32 based hashing candidates use the CRC32
algorithm they will have the same distribution. It is sufficient to benchmark the `crc32opt` algorithm alone.

|Candidate|n|Distribution factor|
|---|---|---|
|crc32|10|0.6|
|djb2|10|1.0|

*I was not able to measure the distribution factor of the `goh` candidate*.

## Discussion

The Go language built-in hash map outperforms both CRC32 hash table implementations. It sees the greatest performance
advantage in *lookup* operations, which is common in a symbol table setting. The Go hash map experiences less speedup
in *remove* operation compared to one of the CRC32 implementations, `crc32re`, but it is still considerably faster. When
we look at the
*lookup* benchmark, by far the most performance critical property of a symbol table, we see that the built-in Go hash
map is more than twice as performant as three other candidates. It seems that even the optimised
`crc32opt`, which works in similar fashion to `goh` using bucket tables, and using pre-defined data types does not
considerably improve *lookup* times than the other two CRC32 candidates.

What causes the performance differences is beyond the scope of this project, but here are some ideas on what might be
going on.

- Buckets: The CRC32 candidates `crc32p` and `crc32re` uses a slice and a linked list per index. The Go hashmap uses a
  hash table of buckets, where each bucket is an [8 pair table](https://go.dev/src/runtime/map.go). If the bucket is
  full and needs to hold more elements a new equally sized bucket is chained/linked to the full bucket. The Go hashmap
  is a mix of linked lists and bucket tables. `crc32opt` uses the same bucket system; and it does not considerably
  improve *lookup* performance. It seems that buckets speed up *insertion*, but slows down *removal*. For symbol table
  usage it is more preferable to have faster insertions.
- The Go hash
  map [generates](https://dave.cheney.net/2018/05/29/how-the-go-runtime-implements-maps-efficiently-without-generics)
  an optimised hashmap at compile time. The CRC32 implementations `crc32p` and `crc32re` uses `interface{}` to
  accommodate arbitrary data types. `crc32opt` only accepts fixed data types, but does not considerably speed up *
  lookup*.
- The Go hash map uses a different hashing algorithm. It has been
  [suggested](https://stackoverflow.com/questions/37625480/how-does-go-calculate-a-hash-value-for-keys-in-a-map)
  that Go might use an AES based algorithm for its hashmap, but this *could change* over time.
- The benchmark was run in-order from file top to bottom. There could be discrepancies and unfavourable memory alignment
  as a result of testing the entire file from top to bottom which favours one or the other candidate.
- The implementation and port of the CRC32 algorithm was my own doing, and it is very likely that the implementors of
  the Go hashmap were capable of creating a better performing hashing solution than yours truly.
- The CRC32 algorithm has a distribution factor of 0.6 during my tests. This means that 40% of test data was hashed to
  an index that a previous data element was hashed to. For the
  `crc32p` and `crc32re` candidates this incurs costs of traversing linked lists, while `crc32opt`
  incurs a linear slice search. However, performance didn't improve drastically when using the DJB2 algorithm, which
  boasts a perfect 1.0 distribution factor in my tests.

## Conclusion

The Go hashmap has proven to be better performing than general data type CRC32 and DJB2 implementations presented in
this repository. It is more than twice as performant in the most crucial task: *lookups*. It supports arbitrary types 
while omitting the need for generic data types such as `interface{}`. It has an easy-to-use interface for lookup and
map initiation using array indexer and `make` function.
